.486                  ; the instruction set we want the compiler to use
.model flat, stdcall  ; 32bit flat memory addressing / stdcall proc calling
option casemap :none  ; case sensative names

;===========================================================================
; *** if you dont have MASM installed at C:\MASM you will have to adjust 
; the paths on these include statements to describe the correct location
;===========================================================================


; -- equates for a big pile of windows constants
include    C:\masm32\include\windows.inc 

; -- libs and prototypes for importing API functions
; -- from kernel32.dll and user32.dll
include    C:\masm32\include\kernel32.inc 
includelib C:\masm32\lib\kernel32.lib
include    C:\masm32\include\user32.inc 
includelib C:\masm32\lib\user32.lib



; MACROs are just shorthand ways to write things, their correct
; equivalents are substituted into the SOURCE code before it is compiled

; always a handy macro ;
m2m MACRO m1, m2
    push m2
    pop  m1
ENDM



; the "equ" keyword defines "equates", these are constants that will
; end up as a single numeric value when the program is compiled, allthough
; they can be defined using logic and/or mathematical operators, and can 
; be the product of other equates.

CLASS_STYLE             equ CS_VREDRAW+CS_HREDRAW
WINSTYLE                equ WS_OVERLAPPEDWINDOW
WINEXSTYLE              equ 0
POPUP_FLAGS             equ MF_BYPOSITION OR MF_POPUP

; invent some arbitrary ID numbers for our menu items

MenuID_ABOUT            equ WM_USER+201h
MenuID_EXIT             equ WM_USER+202h


;       the ".data?" section contains only uninitialized data, so everything
;       here must be declared as "?" (undefined)
.data?

hWindow             dd  ?     ; "dd" = "declare double" this means allocate
hMenu               dd  ?     ; 4 bytes for a DWORD
hMenuFile           dd  ?     

; Windows Message Structure
  MSGhWnd           dd  ?
  MSGmessage        dd  ?
  MSGwParam         dd  ?
  MSGlParam         dd  ?
  MSGtime           dd  ?
  MSGptX            dd  ?
  MSBptY            dd  ?

; the ".data" section - initialized data
.data

AppName             db  "PUDed V0.0.1a",0 ; "db" = "declare byte"
					  ; this means allocate 1 byte for
					  ; each of these (i.e.an ANSII string)
ClassName           db  "PUDed_Class",0
				    ; the ".data" section contains initialized
szMenuFile          db  "&File",0   ; data, so what is defined here will be
szMenuAbout         db  "&About",0  ; what is at these addresses when the app
szMenuExit          db  "E&xit",0   ; first executes, although they can be
				    ; changed/updated as the program runs.


winXpos             dd  150
winYpos             dd  150

winWidth            dd  640
winHeight           dd  480


AboutText           db  "PUDed - Warcraft II Scenario File editor",0Ah,0Dh
		    db  0Ah,0Dh
		    db  "         A collaborative project from",0Ah,0Dh
		    db  " the die-hard WC2 players on war2.ru",0Ah,0Dh,0





    ;   Windows WNDCLASSEX Structure
    ;   ( defines a window class )

    CEXcbSize            dd      30h ; Size of WNDCLASSEX structure (48 bytes)
    CEXstyle             dd      CLASS_STYLE ; class style defined as an equate
    CEXlpfnWndProc       dd      OFFSET WndProc ; address of our WndProc func
    CEXcbClsExtra        dd      0  ; no extra allocated for class
    CEXcbWndExtra        dd      0  ; no extra allocated for instance
    CEXhInstance         dd      0  ;* Need to fill the instance handle here *
    CEXhIcon             dd      0  ; Default icon is fine for now
    CEXhCursor           dd      0  ;* Need to fill a cursor handle here *
    CEXhbrBackground     dd      10h; standard background brush
    CEXlpszMenuName      dd      0  ; no default menu for the class
    CEXlpszClassName     dd      OFFSET ClassName
    CEXhIconSm           dd      0  ; Default icon is fine for now



; the ".code" section contains the intsructions that will be sent to the CPU
; when the program executes. (this section is usually designated ".text"
; internally when the .exe is linked, however the MASM syntax desigtnates it
; as ".code" here.)

.code

; "WndProc" is part of the Windows operating system structure.
;
; Every window class must have a WndProc so it can interact with the user and
; the OS. Whenever an OS event needs to interact with our window, a message is
; sent to the to this procedure so we can decide what to do with the message.
;
; If it's not a message we need, then we just pass it on to "DefWindowProc"
; which handles the default message processing for standard windows events.
;
; i.e. if the user pushes a button on our app, it will generate a WM_COMMAND
; message, then our WndProc will be called with that message, which contains
; the ID of the button that was pressed. If we are displaying graphics on our
; window, then we will want to handle WM_PAINT messages, which are generated
; whenever part of the window needs to be redrawn. If we want to keep track of
; how big the window is, then we must handle WM_SIZE messages, which are sent
; whenever the user resizes the window. etc.

WndProc proc WPhWnd   :DWORD, ;   \
	     WPuMsg   :DWORD, ;    \ WndProc is passed 4 DWORD arguments
	     WPwParam :DWORD, ;    /
	     WPlParam :DWORD  ;   /

  ; is it a "command"? (i.e. a push-button or menu selection)
  
  cmp WPuMsg, WM_COMMAND
  ; if not then skip this bit
  jnz NOT_COMMAND
  ;         =============================
  ;         >>>>>> COMMAND MESSAGES <<<<<
  ;         =============================
  
    ; if it is a WM_COMMAND message then wParam holds the ID of 
    ; the control that generated it
    
    cmp WPwParam,MenuID_EXIT
    jnz  @F
      ;"Exit" selected from menu, shut down the app
      push 0
      call PostQuitMessage
      ret
    @@:
    
    cmp WPwParam,MenuID_ABOUT
    jnz  @F
      ;"About" selected from menu, display a simple message box.
      push 0
      push OFFSET AppName
      push OFFSET AboutText
      push hWindow
      call MessageBox
      ret
    @@:
    ;unknown or unhandled control ID
    ret
  
  NOT_COMMAND:
  ;std windows messages
  cmp WPuMsg, WM_CREATE
  jnz @F
    ; this is called when a window is created, 
    ; some initialization code can go here
    ret    
  @@:
    ; We will want to handle WM_PAINT messages later,
    ; but for now we can just leave this to DefWindowProc
    ;
    ;cmp WPuMsg, WM_PAINT
    ;jnz @F
    ;ret    
  ;@@:
    
  cmp WPuMsg, WM_DESTROY
    ; window has been closed, shut down the app
    jnz @F
    push NULL
    call PostQuitMessage
    ret                    
  @@:
  
  ; Made it all the way here, so its not a message we are handling.
  ; we just pass it along to DefWindowProc
  push WPlParam
  push WPwParam
  push WPuMsg
  push WPhWnd
  call DefWindowProc
  ret
WndProc endp





start:      ; "start" defines where the app entry point will be.
	    ; When the app is initialized, execution will start here.


;==================================
;>>>>>>>>>Register Class<<<<<<<<<<<
;==================================
;
; First we need to register a class for our window.
; Most of the WNDCLASSEX struct is already completed
; in the .data section, there's just a couple of values 
; we need to fill in at run-time.
  
  
; get the instance handle 
  push  0
  call  GetModuleHandle
  mov   CEXhInstance  ,eax

; get a handle to a standard "arrow" pointer
  push  IDC_ARROW
  push  0
  call  LoadCursor
  mov   CEXhCursor, eax

; now we use the WNDCLASSEX structure to register the class
  push  OFFSET CEXcbSize ;   <-- cbSize is the first member of WNDCLASSEX
  call  RegisterClassEx

; no need to save the returned class atom as we can reference
; the class by the name we have registered it with


;   =============================
;   >>>>>>>>Create Menus<<<<<<<<<
;   =============================
; Here we create any menus we want to use with our window
;
; For now we will just make a normal menu that 
; is displayed at the top of our app window

  ; first we create the top level menu object
  call CreateMenu
  mov hMenu,eax
  
  ; then we create the "File" dropdown (aka "popup") menu
  call CreateMenu
  mov hMenuFile,eax
  
  ; add the menu items to the file menu (these are in effect push-buttons)
  
  push OFFSET szMenuAbout ; address of the text "&About"
  push MenuID_ABOUT       ; the ID number we are giving to this command
  push 0                  ; no flags here, just a standard menu item
  push hMenuFile          ; attach it to this menu
  call AppendMenu

  push OFFSET szMenuExit  ; address of the text "E&xit"
  push MenuID_EXIT        ; the ID number we are giving to this command
  push 0                  ; no flags here, just a standard menu item
  push hMenuFile          ; attach it to this menu
  call AppendMenu

  ; add the "File" drop-down menu to the menu object
  
  push OFFSET szMenuFile  ; address of text "&File"
  push hMenuFile          ; handle of drop-down menu
  push POPUP_FLAGS        ; this is a drop-down menu, not a menu item
  push hMenu              ; handle of the menu we are attaching it to
  call AppendMenu        


; Now we can use this menu when we create our window



;           =============================
;           >>>>>>>>Create Window<<<<<<<<
;           =============================
	
	
  push 0                ; don't need to tag WM_CREATE message
  push CEXhInstance     ; already got instance handle
  push hMenu            ; handle of the menu we just made
  push 0                ; no parent - this is our top level window
  push winHeight        ; initial height is set in the .data section
  push winWidth         ; initial width is set in the .data section
  push winYpos          ; initial Y-position is set in the .data section
  push winXpos          ; initial X-position is set in the .data section
  push WINSTYLE         ; our window style defined as an equate above
  push OFFSET AppName   ; text for the window title
  push OFFSET ClassName ; the name of our registered class
  push WINEXSTYLE       ; our window ex-style defined as an equate above
  call CreateWindowEx
	
  ; save the handle to our new window
  mov hWindow,eax

  ; make the window visible
  push 1
  push eax
  call ShowWindow



; OK, that's all our initialization done, now we only need to do things in
; response to user input, so we enter our "message loop" - not to be confused
; with "WndProc".
;
; -- While "WndProc" executes in response to events from the OS, the message
;    loop is what gives control over to the OS in the first place.

; >>The Message Loop<<
; --------------------
; "GetMessage" waits for a message from the OS for this application before
; returning, so in most windows apps, 99% of the time is spent in this proc
; call.. doing nothing, just waiting for the next system event to demand 
; attention. When it is translated and dispatched, the OS will call our 
; WndProc with the message details and we will handle it there.
;
; This may seem like double handling, but there is only one message loop for
; each app, while there may be many different classes of windows (i.e.controls)
; in that one application, with each class having its own "WndProc" to process
; messages correctly for that class, and indeed a single WndProc procedure may
; be used by many simultaneously running processes that are all using instances
; of a particular window class.
;

; examples: 
; ---------
; An EDIT control is actually a class of window that is built in to the OS. 
; When a user types something into one of these, or selects the text and copies
; or pastes etc. The main app window is not notified of every little event, 
; that would just be tedious. The WndProc for the EDIT class which is (I think)
; located somewhere in user32.dll handles all of these events, leaving the 
; parent app free to just check out what string has ended up in it. 

; Similarly a BUTTON control is a class of window that draws itself as a 3D
; looking push-button, and reacts to being clicked by drawing neat graphics that
; simulate a button being pressed when the user clicks and releases the mouse
; button, and perhaps displays a tool-tip if the mouse cursor just hovers over
; the control, but the parent app need not process all of these events. 
; The BUTTON class will send a WM_COMMAND message along with its own ID to is 
; parent window when the user activates it, which is all the app really needs.


message_loop:
; ---------------------------------------------------------------------
; This is the little wheel our hamster runs in at the centre of the app
; ---------------------------------------------------------------------
  
  xor eax,eax
  ; wait for the next message to be sent to our app by the OS
  push eax
  push eax
  push eax
  push OFFSET MSGhWnd
  call GetMessage
  
  test eax,eax
  ; if GetMessage returned 0 the app has terminated
  jz @F
    
    ; >> normal message <<
    
    ; translate it
    push OFFSET MSGhWnd
    call TranslateMessage

    ; dispatch it
    push OFFSET MSGhWnd
    call DispatchMessage
    
    ; go back and wait for the next message
    jmp message_loop
  
  @@: 
  ;>> we are exiting <<
  
push 0
call ExitProcess

ret ; and a ret where you app finally terminates to hand control back to the OS
    ; otherwise it will generate an exception on exit, which is just ugly

end start
